'''
Copyright 2019 Xilinx Inc.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
'''

from ctypes import *
import cv2
import numpy as np
from dnndk import n2cube

try:
    pyc_libdputils = cdll.LoadLibrary("libn2cube.so")
except Exception:
    print('Load libn2cube.so failed\nPlease install DNNDK first!')


def dpuSetInputImageWithScale(task, nodeName, image, mean, scale, idx=0):
    """Set image into DPU Task's input Tensor with a specified scale parameter"""
    height = n2cube.dpuGetInputTensorHeight(task, nodeName, idx)
    width = n2cube.dpuGetInputTensorWidth(task, nodeName, idx)
    channel = n2cube.dpuGetInputTensorChannel(task, nodeName, idx)
    (imageHeight, imageWidth, imageChannel) = image.shape
    inputMean = (c_float * channel)()
    for i in range(0, channel):
        inputMean[i] = mean[i]

    if height == imageHeight and width == imageWidth:
        newImage = image
    else:
        newImage = cv2.resize(image, (width, height), 0, 0, cv2.INTER_LINEAR)

    inputImage = np.asarray(newImage, dtype=np.byte)
    inputImage2 = inputImage.ctypes.data_as(c_char_p)
    return pyc_libdputils.pyc_dpuSetInputData(task,
                                              c_char_p(nodeName.encode("utf-8")), inputImage2,
                                              c_int(height),
                                              c_int(width),
                                              c_int(imageChannel), inputMean,
                                              c_float(scale), c_int(idx))


def dpuSetInputImage(task, nodeName, image, mean, idx=0):
    """
    Set image into DPU Task's input Tensor
    task: DPU Task
    nodeName: The pointer to DPU Node name.
    image:    Input image in OpenCV Mat format. Single channel and 3-channel input image are supported.
    mean:     Mean value array which contains 1 member for single channel input image
              or 3 members for 3-channel input image
              Note: You can get the mean values from the input Caffe prototxt.
                    At present, the format of mean value file is not yet supported
    idx:      The index of a single input tensor for the Node, with default value as 0
    """
    return dpuSetInputImageWithScale(task, nodeName, image, mean, 1.0, idx)


def dpuSetInputImage2(task, nodeName, image, idx=0):
    """
    Set image into DPU Task's input Tensor (mean values automatically processed by N2Cube)
    nodeName: The pointer to DPU Node name.
    image:    Input image in OpenCV Mat format. Single channel and 3-channel input image are supported.
    idx:      The index of a single input tensor for the Node, with default value as 0
    """
    channel = n2cube.dpuGetInputTensorChannel(task, nodeName, idx)
    output = (c_float * channel)()
    outputMean = POINTER(c_float)(output)
    pyc_libdputils.loadMean(task, outputMean, channel)
    for i in range(channel):
        outputMean[i] = float(outputMean[i])
    return dpuSetInputImageWithScale(task, nodeName, image, outputMean, 1.0,
                                     idx)



'''
Copyright 2019 Xilinx Inc.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
'''

import os
import cv2
import colorsys
import random
import numpy as np
from dnndk import n2cube

'''resize image with unchanged aspect ratio using padding'''
def letterbox_image(image, size):
    ih, iw, _ = image.shape
    w, h = size
    scale = min(w/iw, h/ih)
    #print(scale)
    
    nw = int(iw*scale)
    nh = int(ih*scale)
    #print(nw)
    #print(nh)

    image = cv2.resize(image, (nw,nh), interpolation=cv2.INTER_LINEAR)
    new_image = np.ones((h,w,3), np.uint8) * 128
    h_start = (h-nh)//2
    w_start = (w-nw)//2
    new_image[h_start:h_start+nh, w_start:w_start+nw, :] = image
    return new_image

'''image preprocessing'''
def pre_process(image, model_image_size):
    image = image[...,::-1]
    image_h, image_w, _ = image.shape
 
    if model_image_size != (None, None):
        assert model_image_size[0]%32 == 0, 'Multiples of 32 required'
        assert model_image_size[1]%32 == 0, 'Multiples of 32 required'
        boxed_image = letterbox_image(image, tuple(reversed(model_image_size)))
    else:
        new_image_size = (image_w - (image_w % 32), image_h - (image_h % 32))
        boxed_image = letterbox_image(image, new_image_size)
    image_data = np.array(boxed_image, dtype='float32')
    image_data /= 255.
    image_data = np.expand_dims(image_data, 0) 	
    return image_data

'''Get model classification information'''	
def get_class(classes_path):
    with open(classes_path) as f:
        class_names = f.readlines()
    class_names = [c.strip() for c in class_names]
    return class_names

'''Get model anchors value'''
def get_anchors(anchors_path):
    with open(anchors_path) as f:
        anchors = f.readline()
    anchors = [float(x) for x in anchors.split(',')]
    return np.array(anchors).reshape(-1, 2)
    
def _get_feats(feats, anchors, num_classes, input_shape):
    num_anchors = len(anchors)
    anchors_tensor = np.reshape(np.array(anchors, dtype=np.float32), [1, 1, 1, num_anchors, 2])
    grid_size = np.shape(feats)[1:3]
    nu = num_classes + 5
    predictions = np.reshape(feats, [-1, grid_size[0], grid_size[1], num_anchors, nu])
    grid_y = np.tile(np.reshape(np.arange(grid_size[0]), [-1, 1, 1, 1]), [1, grid_size[1], 1, 1])
    grid_x = np.tile(np.reshape(np.arange(grid_size[1]), [1, -1, 1, 1]), [grid_size[0], 1, 1, 1])
    grid = np.concatenate([grid_x, grid_y], axis = -1)
    grid = np.array(grid, dtype=np.float32)

    box_xy = (1/(1+np.exp(-predictions[..., :2])) + grid) / np.array(grid_size[::-1], dtype=np.float32)
    box_wh = np.exp(predictions[..., 2:4]) * anchors_tensor / np.array(input_shape[::-1], dtype=np.float32)
    box_confidence = 1/(1+np.exp(-predictions[..., 4:5]))
    box_class_probs = 1/(1+np.exp(-predictions[..., 5:]))
    return box_xy, box_wh, box_confidence, box_class_probs
	
def correct_boxes(box_xy, box_wh, input_shape, image_shape):
    box_yx = box_xy[..., ::-1]
    box_hw = box_wh[..., ::-1]
    input_shape = np.array(input_shape, dtype = np.float32)
    image_shape = np.array(image_shape, dtype = np.float32)
    new_shape = np.around(image_shape * np.min(input_shape / image_shape))
    offset = (input_shape - new_shape) / 2. / input_shape
    scale = input_shape / new_shape
    box_yx = (box_yx - offset) * scale
    box_hw *= scale

    box_mins = box_yx - (box_hw / 2.)
    box_maxes = box_yx + (box_hw / 2.)
    boxes = np.concatenate([
        box_mins[..., 0:1],
        box_mins[..., 1:2],
        box_maxes[..., 0:1],
        box_maxes[..., 1:2]
    ], axis = -1)
    boxes *= np.concatenate([image_shape, image_shape], axis = -1)
    return boxes
	
def boxes_and_scores(feats, anchors, classes_num, input_shape, image_shape):
    box_xy, box_wh, box_confidence, box_class_probs = _get_feats(feats, anchors, classes_num, input_shape)
    boxes = correct_boxes(box_xy, box_wh, input_shape, image_shape)
    boxes = np.reshape(boxes, [-1, 4])
    box_scores = box_confidence * box_class_probs
    box_scores = np.reshape(box_scores, [-1, classes_num])
    return boxes, box_scores	

'''Draw detection frame'''
def draw_bbox(image, bboxes, classes):
    """
    bboxes: [x_min, y_min, x_max, y_max, probability, cls_id] format coordinates.
    """
    num_classes = len(classes)
    image_h, image_w, _ = image.shape
    hsv_tuples = [(1.0 * x / num_classes, 1., 1.) for x in range(num_classes)]
    colors = list(map(lambda x: colorsys.hsv_to_rgb(*x), hsv_tuples))
    colors = list(map(lambda x: (int(x[0] * 255), int(x[1] * 255), int(x[2] * 255)), colors))

    random.seed(0)
    random.shuffle(colors)
    random.seed(None)

    for i, bbox in enumerate(bboxes):
        coor = np.array(bbox[:4], dtype=np.int32)
        fontScale = 0.5
        score = bbox[4]
        class_ind = int(bbox[5])
        bbox_color = colors[class_ind]
        bbox_thick = int(0.6 * (image_h + image_w) / 600)
        c1, c2 = (coor[0], coor[1]), (coor[2], coor[3])
        cv2.rectangle(image, c1, c2, bbox_color, bbox_thick)
    return image
	
def nms_boxes(boxes, scores):
    """Suppress non-maximal boxes.

    # Arguments
        boxes: ndarray, boxes of objects.
        scores: ndarray, scores of objects.

    # Returns
        keep: ndarray, index of effective boxes.
    """
    x1 = boxes[:, 0]
    y1 = boxes[:, 1]
    x2 = boxes[:, 2]
    y2 = boxes[:, 3]

    areas = (x2-x1+1)*(y2-y1+1)
    order = scores.argsort()[::-1]

    keep = []
    while order.size > 0:
        i = order[0]
        keep.append(i)

        xx1 = np.maximum(x1[i], x1[order[1:]])
        yy1 = np.maximum(y1[i], y1[order[1:]])
        xx2 = np.minimum(x2[i], x2[order[1:]])
        yy2 = np.minimum(y2[i], y2[order[1:]])

        w1 = np.maximum(0.0, xx2 - xx1 + 1)
        h1 = np.maximum(0.0, yy2 - yy1 + 1)
        inter = w1 * h1

        ovr = inter / (areas[i] + areas[order[1:]] - inter)
        inds = np.where(ovr <= 0.55)[0]  # threshold
        order = order[inds + 1]

    return keep	
  
'''Model post-processing'''
def eval(yolo_outputs, image_shape, max_boxes = 20):
    score_thresh = 0.2
    nms_thresh = 0.45
    class_names = get_class(classes_path)
    anchors     = get_anchors(anchors_path)
    anchor_mask = [[6, 7, 8], [3, 4, 5], [0, 1, 2]]
    boxes = []
    box_scores = []
    
    input_shape = np.shape(yolo_outputs[0])[1 : 3]
    #print(input_shape)
    input_shape = np.array(input_shape)*32
    #print(input_shape)
    
    for i in range(len(yolo_outputs)):
        _boxes, _box_scores = boxes_and_scores(yolo_outputs[i], anchors[anchor_mask[i]], len(class_names), input_shape, image_shape)
        boxes.append(_boxes)
        box_scores.append(_box_scores)
    boxes = np.concatenate(boxes, axis = 0)
    box_scores = np.concatenate(box_scores, axis = 0)
    
    mask = box_scores >= score_thresh
    boxes_ = []
    scores_ = []
    classes_ = []
    for c in range(len(class_names)):
        class_boxes_np = boxes[mask[:, c]]
        class_box_scores_np = box_scores[:, c]
        class_box_scores_np = class_box_scores_np[mask[:, c]]
        nms_index_np = nms_boxes(class_boxes_np, class_box_scores_np) 
        class_boxes_np = class_boxes_np[nms_index_np]
        class_box_scores_np = class_box_scores_np[nms_index_np]
        classes_np = np.ones_like(class_box_scores_np, dtype = np.int32) * c
        boxes_.append(class_boxes_np)
        scores_.append(class_box_scores_np)
        classes_.append(classes_np)
    boxes_ = np.concatenate(boxes_, axis = 0)
    scores_ = np.concatenate(scores_, axis = 0)
    classes_ = np.concatenate(classes_, axis = 0)
    
    return boxes_, scores_, classes_

"""DPU Kernel Name for tf_yolov3_voc"""
KERNEL_CONV="tf_yolov3_voc"

"""DPU IN/OUT Name for tf_yolov3_voc"""
CONV_INPUT_NODE="conv2d_1_convolution"
CONV_OUTPUT_NODE1="conv2d_59_convolution"
CONV_OUTPUT_NODE2="conv2d_67_convolution"
CONV_OUTPUT_NODE3="conv2d_75_convolution"

if __name__ == "__main__":

    """ Attach to DPU driver and prepare for running """
    n2cube.dpuOpen()

    """ Create DPU Kernels for tf_yolov3_voc """
    kernel = n2cube.dpuLoadKernel(KERNEL_CONV)

    """ Create DPU Tasks for tf_yolov3_voc """
    task = n2cube.dpuCreateTask(kernel, 0)
	
    """Load image to DPU"""
    image_path = "./image/input.jpg"
    #print("Loading  image...")
    image = cv2.imread(image_path)	
    image_ho, image_wo, _ = image.shape
    image_size = image.shape[:2]
    image_data = pre_process(image, (416, 416))
    
    image_data = np.array(image_data,dtype=np.float32)
    input_len = n2cube.dpuGetInputTensorSize(task, CONV_INPUT_NODE)
 
    """Get input Tesor"""
    n2cube.dpuSetInputTensorInHWCFP32(task,CONV_INPUT_NODE,image_data,input_len)

    """Model run on DPU"""
    n2cube.dpuRunTask(task)
   
    """Get the output tensor"""   
    conv_sbbox_size = n2cube.dpuGetOutputTensorSize(task, CONV_OUTPUT_NODE1)
    conv_out1 = n2cube.dpuGetOutputTensorInHWCFP32(task, CONV_OUTPUT_NODE1, conv_sbbox_size)
    conv_out1 = np.reshape(conv_out1, (1, 13, 13, 75))
    
    conv_mbbox_size = n2cube.dpuGetOutputTensorSize(task, CONV_OUTPUT_NODE2)
    conv_out2 = n2cube.dpuGetOutputTensorInHWCFP32(task, CONV_OUTPUT_NODE2, conv_mbbox_size)
    conv_out2 = np.reshape(conv_out2, (1, 26, 26, 75))
        
    conv_lbbox_size = n2cube.dpuGetOutputTensorSize(task, CONV_OUTPUT_NODE3)
    conv_out3 = n2cube.dpuGetOutputTensorInHWCFP32(task, CONV_OUTPUT_NODE3, conv_lbbox_size)
    conv_out3 = np.reshape(conv_out3, (1, 52, 52, 75))
    
    yolo_outputs = [conv_out1, conv_out2, conv_out3]    
          
    """Get model classification information"""
    classes_path = "./image/voc_classes.txt"
    class_names = get_class(classes_path)
    
    """Get model anchor value"""
    anchors_path = "./model_data/yolo_anchors.txt"
    anchors = get_anchors(anchors_path)
    
    """Post-processing"""   
    out_boxes, out_scores, out_classes = eval(yolo_outputs, image_size)
	
    items = []
    draws = []
    for i, c in reversed(list(enumerate(out_classes))):
        predicted_class = class_names[c]
        box = out_boxes[i]
        score = out_scores[i]

        top, left, bottom, right = box
        top = max(0, np.floor(top + 0.5).astype('int32'))
        left = max(0, np.floor(left + 0.5).astype('int32'))
        bottom = min(image_ho, np.floor(bottom + 0.5).astype('int32'))
        right = min(image_wo, np.floor(right + 0.5).astype('int32'))
        draw  = [left, top, right, bottom, score, c]
        item  = [predicted_class, score, left, top, right, bottom]
        draws.append(draw)
        items.append(item)
	
    image_id = 0
    image_result = draw_bbox(image, draws, class_names)
    cv2.imshow("showing",image_result)
    cv2.waitKey(0)
    cv2.destroyAllWindows()



#  Copyright (C) 2020 Xilinx, Inc
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.


from .dpu import DpuOverlay


__author__ = "Yun Rock Qu, Jingwei Zhang"
__copyright__ = "Copyright 2020, Xilinx"
__email__ = "pynq_support@xilinx.com"
__version__ = '1.2.0'



#  Copyright (C) 2020 Xilinx, Inc
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

__author__ = "Yun Rock Qu, Jingwei Zhang"
__copyright__ = "Copyright 2020, Xilinx"
__email__ = "pynq_support@xilinx.com"


from setuptools import setup, find_packages, Distribution
from setuptools.command.build_ext import build_ext
import shutil
import os
import platform
import re
import warnings


# temporary fix to ignore "Device Not Found" error
try:
    from pynq.utils import build_py
except RuntimeError:
    pass


# global variables
module_name = "pynq_dpu"
git_submodule_vai = "vitis-ai-git"
data_files = []
dnndk_applications = {'resnet50': "dpu_resnet50_0.elf",
                      'inception_v1': "dpu_inceptionv1_0.elf",
                      'tf_yolov3_voc_py': "dpu_tf_yolov3_0.elf"}


# parse version number
def find_version(file_path):
    with open(file_path, 'r') as fp:
        version_file = fp.read()
        version_match = re.search(r"^__version__ = ['\"]([^'\"]*)['\"]",  
                                  version_file, re.M)
    if version_match:
        return version_match.group(1)
    raise NameError("Version string must be defined in {}.".format(file_path))


# extend package
def extend_package(path):
    if os.path.isdir(path):
        data_files.extend(
            [os.path.join("..", root, f)
             for root, _, files in os.walk(path) for f in files]
        )
    elif os.path.isfile(path):
        data_files.append(os.path.join("..", path))


# merge 2 folder recursively
def copy_tree(root_src_dir, root_dst_dir):
    for src_dir, dirs, files in os.walk(root_src_dir):
        dst_dir = src_dir.replace(root_src_dir, root_dst_dir, 1)
        if not os.path.exists(dst_dir):
            os.makedirs(dst_dir)
        for f in files:
            src_file = os.path.join(src_dir, f)
            dst_file = os.path.join(dst_dir, f)
            if os.path.exists(dst_file):
                os.remove(dst_file)
            shutil.copy(src_file, dst_dir)


# copy DNNDK applications
def copy_dnndk_samples(src, dst):
    """Copy DNNDK samples.

    We will only copy the samples that we are interested. More samples can be
    added to the `application` list. We also need to refrain from storing
    large ML models in source distribution; hence we need to remove all the
    models.

    We will just rely on downloading large ML models from Xilinx opendownload
    when installing the package; at that time we will download specific ones
    depending on the BOARD environment.

    """
    copy_tree(os.path.join(src, 'common'),
              os.path.join(dst, 'common'))
    os.remove(os.path.join(dst, 'common', 'dputils.py'))
    for app in dnndk_applications:
        src_app = os.path.join(src, app)
        dst_app = os.path.join(dst, app)
        copy_tree(src_app, dst_app)
        os.system("cd {} && ".format(dst_app) +
                  "rm -rf build.sh Makefile model_for_*")


# Customized patch
def patch_ignore_prints(file_path):
    """Patch Python file to ignore prints.

    Some Python files in DNNDK examples have too many distracting print outs.
    This helper function will patch the Python file so we can ignore print outs.

    """
    os.system('sed -i "s/print/#print/g" {}'.format(file_path))


# Enforce platform-dependent distribution
class CustomDistribution(Distribution):
    def has_ext_modules(self):
        return True


# get current platform: edge
def get_platform():
    cpu = platform.processor()
    if cpu in ['armv7l', 'aarch64']:
        return "edge"
    elif cpu in ['x86_64']:
        raise OSError("Platform is not supported.")
    else:
        raise OSError("Platform is not supported.")


# get current board
def get_board():
    board_variable = os.getenv('BOARD')
    if board_variable not in ['Ultra96', 'ZCU104', 'ZCU111']:
        warnings.warn(UserWarning, "Board is not officially supported.")
    return board_variable.lower().replace('-', '')


# install dnndk package
def install_dnndk_pkg(pkg_path):
    os.system('cd {} && '
              'wget -O vitis-ai_v1.2_dnndk.pynq.tar.gz '
              '"https://www.xilinx.com/bin/public/openDownload?filename='
              'vitis-ai_v1.2_dnndk.pynq.tar.gz" && '
              'tar -xvf vitis-ai_v1.2_dnndk.pynq.tar.gz && '
              'cd vitis-ai_v1.2_dnndk && '
              'bash install.sh'.format(pkg_path))


# install vart package

def install_vart_pkg(pkg_path, edge):
    os.system('cd {0} && '
              'wget -O vitis-ai-runtime-1.2.pynq.tar.gz '
              '"https://www.xilinx.com/bin/public/openDownload?filename='
              'vitis-ai-runtime-1.2.pynq.tar.gz" && '
              'tar -xvf vitis-ai-runtime-1.2.pynq.tar.gz && '
              'cd {1} && '
              'apt-get install ./*.deb && '
              'cd ../ && '
              'rm -rf *.tar.gz aarch64 armv7l'.format(pkg_path, edge))


# resolve overlay files by moving the cached copy
def resolve_overlay_d(path):
    subfolders = [os.path.abspath(f.path) for f in os.scandir(path)
                  if f.is_dir() and f.path.endswith('.d')]
    for f in subfolders:
        dst_name = f.rstrip('.d')
        file_list = [os.path.join(f, x) for x in os.listdir(f)]
        if not file_list:
            raise FileNotFoundError("Folder {} is empty.".format(f))
        shutil.copy(file_list[0], dst_name)
        shutil.rmtree(f)


class BuildExtension(build_ext):
    """A custom build extension for adding user-specific options.

    This will first install DNNDK package targeting the current board.
    Also, a couple of link files will be resolved at the end so users
    have the correct overlay files in the `overlays` folder.

    """
    def run(self):
        pkg_path = os.path.join(module_name, get_platform())
        install_dnndk_pkg(pkg_path)
        install_vart_pkg(pkg_path, platform.processor())
        build_ext.run(self)
        overlay_path = os.path.join(self.build_lib, module_name, 'overlays')
        resolve_overlay_d(overlay_path)


pkg_version = find_version('{}/__init__.py'.format(module_name))
with open("README.md", encoding='utf-8') as fh:
    readme_lines = fh.readlines()[2:7]
long_description = (''.join(readme_lines))


if os.path.isdir(git_submodule_vai):
    src_img = os.path.join(git_submodule_vai,
                           'DPU-TRD/app/img')
    dst_img = os.path.join(module_name, 'edge/notebooks/img')
    dst_data = os.path.join(module_name, 'edge/notebooks/data')
    src_dnndk_samples = os.path.join(git_submodule_vai,
                                     'mpsoc/vitis_ai_dnndk_samples')
    dst_dnndk_samples = os.path.join(module_name,
                                     'edge/dnndk')

    copy_tree(src_img, dst_img)
    shutil.copy2(os.path.join(src_dnndk_samples,
                              'inception_v1_mt/image/words.txt'),
                 dst_img)
    shutil.copy2(os.path.join(src_dnndk_samples,
                              'tf_yolov3_voc_py/image/voc_classes.txt'),
                 dst_img)
    shutil.copy2(os.path.join(src_dnndk_samples, 'common/dputils.py'),
                 module_name)
    copy_dnndk_samples(src_dnndk_samples, dst_dnndk_samples)
    patch_ignore_prints(os.path.join(dst_dnndk_samples,
                        'tf_yolov3_voc_py/tf_yolov3_voc.py'))


extend_package(os.path.join(module_name, 'overlays'))
for arch in ['edge']:
    extend_package(os.path.join(module_name, arch))


setup(
    name=module_name,
    version=pkg_version,
    description="PYNQ DPU Overlay using Vitis AI",
    long_description=long_description,
    long_description_content_type='text/markdown',
    author='Xilinx PYNQ Development Team',
    author_email="pynq_support@xilinx.com",
    url='https://github.com/Xilinx/DPU-PYNQ.git',
    license='Apache License 2.0',
    packages=find_packages(),
    package_data={
        "": data_files,
    },
    python_requires=">=3.6.0",
    install_requires=[
        "pynq>=2.6.0",
        "pybind11",
        "CppHeaderParser",
        "mnist"
    ],
    entry_points={
        "pynq.notebooks": [
            "pynq-dpu = {}.{}.notebooks".format(
                module_name, get_platform())
        ]
    },
    cmdclass={"build_py": build_py,
              "build_ext": BuildExtension},
    distclass=CustomDistribution
)



#  Copyright (C) 2020 Xilinx, Inc
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.


import os
import subprocess
import pathlib
import pynq
from pynq import Register
import runner
import xir.graph
import xir.subgraph


__author__ = "Yun Rock Qu, Jingwei Zhang"
__copyright__ = "Copyright 2020, Xilinx"
__email__ = "pynq_support@xilinx.com"


MODULE_PATH = os.path.dirname(os.path.realpath(__file__))
OVERLAY_PATH = os.path.join(MODULE_PATH, 'overlays')
XCL_DST_PATH = "/usr/lib"


def get_kernel_name_for_dnndk(model):
    """Parse the kernel name out of the given elf file for DNNDK runtime.

    This method will take elf file as input, read its information, and
    return the model name.

    Parameters
    ----------
    model : str
        The name of the ML model binary. Can be absolute or relative path.

    """
    cmd = 'readelf {} -s --wide | grep ' \
        '"0000000000000000     0 FILE    LOCAL  DEFAULT  ABS  "'.format(model)
    line = subprocess.check_output(cmd, shell=True).decode()
    kernel_0 = line.split()[-1].lstrip("dpu_").rstrip(".s")
    return kernel_0[:-len('_0')] if kernel_0.endswith('_0') else kernel_0


def get_kernel_name_for_vart(model):
    """Parse the kernel name out of the given elf file for VART.

    This method will leverage the command `dpu_model_inspect` to
    return the model name.

    Parameters
    ----------
    model : str
        The name of the ML model binary. Can be absolute or relative path.

    Returns
    -------
    (str, str)
        Model name and kernel name of the given elf file.

    """
    cmd = "dpu_model_inspect {} | grep 'model_name'".format(model)
    line0 = subprocess.check_output(cmd, shell=True).decode()
    cmd = "dpu_model_inspect {} | grep 'kernel\\[0\\]'".format(model)
    line1 = subprocess.check_output(cmd, shell=True).decode()
    return line0.split('=')[-1].lstrip().rstrip(),\
        line1.split('=')[-1].lstrip().rstrip()

def get_subgraph(g):
    sub = []
    root = g.get_root_subgraph()
    
    sub = [s for s in root.children
            if s.metadata.get_attr_str("device") == "DPU"]
    return sub

class DpuOverlay(pynq.Overlay):
    """DPU overlay class.

    This class inherits from the PYNQ overlay class. The initialization method
    is similar except that we have additional bit file search path.

    """
    def __init__(self, bitfile_name, dtbo=None,
                 download=True, ignore_version=False, device=None):
        """Initialization method.

        Check PYNQ overlay class for more information on parameters.

        By default, the bit file will be searched in the following paths:
        (1) the `overlays` folder inside this module; (2) an absolute path;
        (3) the relative path of the current working directory.

        By default, this class will set the runtime to be `dnndk`.

        """
        if os.path.isfile(bitfile_name):
            abs_bitfile_name = bitfile_name
        elif os.path.isfile(os.path.join(OVERLAY_PATH, bitfile_name)):
            abs_bitfile_name = os.path.join(OVERLAY_PATH, bitfile_name)
        else:
            raise FileNotFoundError('Cannot find {}.'.format(bitfile_name))
        super().__init__(abs_bitfile_name,
                         dtbo=dtbo,
                         download=download,
                         ignore_version=ignore_version,
                         device=device)
        self.overlay_dirname = os.path.dirname(self.bitfile_name)
        self.overlay_basename = os.path.basename(self.bitfile_name)
        self.runtime = 'dnndk'
        self.runner = None
        self.graph = None

    def download(self):
        """Download the overlay.

        This method overwrites the existing `download()` method defined in
        the overlay class. It will download the bitstream, set AXI data width,
        copy xclbin and ML model files.

        """
        super().download()
        self.overlay_dirname = os.path.dirname(self.bitfile_name)
        self.overlay_basename = os.path.basename(self.bitfile_name)

        self.copy_xclbin()

    def set_runtime(self, runtime):
        """Set runtime for the DPU.

        Parameters
        ----------
        runtime: str
            Can be either `dnndk` or `vart`.

        """
        if runtime not in ['dnndk', 'vart']:
            raise ValueError('Runtime can only be dnndk or vart.')
        self.runtime = runtime

    def copy_xclbin(self):
        """Copy the xclbin file to a specific location.

        This method will copy the xclbin file into the destination directory to
        make sure DNNDK libraries can work without problems.

        The xclbin file, if not set explicitly, is required to be located
        in the same folder as the bitstream and hwh files.

        The destination folder by default is `/usr/lib`.

        """
        abs_xclbin = self.overlay_dirname + "/" + \
            self.overlay_basename.rstrip(".bit") + ".xclbin"
        if not os.path.isfile(abs_xclbin):
            raise ValueError(
                "File {} does not exist.".format(abs_xclbin))

        if not os.path.isdir(XCL_DST_PATH):
            raise ValueError(
                "Folder {} does not exist.".format(XCL_DST_PATH))
        _ = subprocess.check_output(["cp", "-f",
                                     abs_xclbin, XCL_DST_PATH])

    def load_model(self, model):
        """Load DPU models for both DNNDK runtime and VART.

        For DNNDK, this method will compile the ML model `*.elf` binary file,
        compile it into `*.so` file located in the destination directory
        on the target. This will make sure DNNDK libraries can work
        without problems.

        The ML model file, if not set explicitly, is required to be located
        in the same folder as the bitstream and hwh files.

        The destination folder by default is `/usr/lib`.

        Currently only `*.elf` files are supported as models. The reason is
        that `*.so` usually have to be recompiled targeting a specific
        rootfs.

        For VART, this method will automatically generate the `meta.json` file
        in the same folder as the model file.

        Parameters
        ----------
        model : str
            The name of the ML model binary. Can be absolute or relative path.

        """
        if os.path.isfile(model):
            abs_model = model
        elif os.path.isfile(self.overlay_dirname + "/" + model):
            abs_model = self.overlay_dirname + "/" + model
        else:
            raise ValueError(
                "File {} does not exist.".format(model))
        if not os.path.isdir(XCL_DST_PATH):
            raise ValueError(
                "Folder {} does not exist.".format(XCL_DST_PATH))

        if not model.endswith(".elf"):
            raise RuntimeError("Currently only elf files can be loaded.")
        else:
            if self.runtime == 'dnndk':
                kernel_name = get_kernel_name_for_dnndk(abs_model)
                model_so = "libdpumodel{}.so".format(kernel_name)
                _ = subprocess.check_output(
                    ["gcc", "-fPIC", "-shared", abs_model, "-o",
                     os.path.join(XCL_DST_PATH, model_so)])
            elif self.runtime == 'vart':
                model_name, kernel_name = get_kernel_name_for_vart(abs_model)
                self.graph = xir.graph.Graph.deserialize(pathlib.Path(abs_model))   
                subgraphs = get_subgraph(self.graph)
                assert len(subgraphs) == 1
                self.runner = runner.Runner(subgraphs[0],"run")
            else:
                raise ValueError('Runtime can only be dnndk or vart.')
